#include "SubPic.h"
#include "Cell.h"

#define min(A,B)((A < B)?A:B)
#define max(A,B)((A > B)?A:B)

#define MAX_WIDTH 800
#define MAX_HEIGHT 600


static uchar PixelArray[MAX_WIDTH*MAX_HEIGHT];
uchar*	CSubPic::m_Pixels;


/////////////////////////////////////////////////////////////////
int CSubPic::m_DisplayHeight = 576;

CSubPic::CSubPic()
{
	Reset();
	m_LastSetPixel = new ushort[MAX_HEIGHT];
	m_StartOfLine = new ushort[MAX_HEIGHT];
	m_Data = 0;
	m_Pixels = PixelArray;
}
	

CSubPic::~CSubPic()
{
	delete[] m_Data;
	delete[] m_LastSetPixel;
	delete[] m_StartOfLine;
}

void CSubPic::Reset(bool ResetAll)
{
	m_NavPackLBA = 0;
	m_CurrentBit = 0;		
	m_CurrentByte = 0;	
	m_PixelCount = 0;	
	m_CurrentPixelX = 0;	
	m_CurrentPixelY = 0;	
	m_SubPicWidth = 0;		
	m_SubPicHeight = 0;
	m_ModSubPicWidth = 0;
	m_SubPicSize = 0;
	m_DataBytesRead = 0;
	m_DataBytesWritten = 0;
	m_TopMostPixel = -1;
	m_BottomMostPixel = -1;
	m_BottomMostPixelT = -1;
	m_LastVerticalMove = 0;
	m_BackgroundPix = -1;
	m_HasButtons = 0;
	m_VerticalBorders = 0;

	m_StartX = 0;
	m_StartY = m_StartYOrig = 0;
	m_EndX = 0;
	m_EndY = m_EndYOrig = 0;
	m_TopFieldOffset = 0;
	m_BotFieldOffset = 0;
	m_EndOfField = 0;

	memset(m_Colors,0,4);
	memset(m_Alphas,0,4);
	m_e2Color = 0;
	m_e1Color = 0;
	m_pColor = 0;
	m_bColor = 0;
	m_e2Alpha = 0;
	m_e1Alpha = 0;
	m_pAlpha = 0;
	m_bAlpha = 0;

	m_SPUReadOK = 0;
	m_SubPicDecoded = 0;
	m_RightMostPixel = 0;
	m_NumPackInSPU = 0;
	memset(m_FilesFrom,0,32*sizeof(FILE*));
	memset(m_LBAFrom,0,32*sizeof(int));

	m_Selected = 0;
	m_Modified = 0;
	m_VOBID = 0;
	m_CELLID = 0;
	m_StreamID = 0;
	m_SPUNumber = 0;
}


uchar CSubPic::pop2()
{
	assert(m_CurrentBit >= 1);
	uchar val = m_Data[m_CurrentByte];
	// Move val all the way to the right, then mask, keeping last 2 bits.
	val >>= (m_CurrentBit-1);
	val &= 0x03;
	// Decrementn m_CurrentBit, if needed go into next byte.
	m_CurrentBit -= 2;
	if(m_CurrentBit == -1)
	{
		m_CurrentBit = 7;
		m_CurrentByte ++;
	}
	assert(m_CurrentByte <= m_EndOfField);
	return val;
}

uchar CSubPic::pop4()
{
	// This will be a bit slow, but it's much much easier this way!
	uchar val = pop2();
	return val*4+pop2();
}

uchar CSubPic::pop8()
{
	// This will be a bit slow, but it's much much easier this way!
	uchar val = pop4();
	return val*16+pop4();
}

int CSubPic::Out(int n, int c)
{
	int i;

	if(n == 0)
	{
		// Special case: output until end of line.
		n = m_SubPicWidth-m_CurrentPixelX;
		m_LastSetPixel[m_CurrentPixelY] = m_CurrentPixelX;
		m_RightMostPixel = max(m_RightMostPixel,m_CurrentPixelX);
	}

	// The background pixel type is the first one we see. This should work most of the time, but not always.
	if(!m_SubPicDecoded && m_PixelCount == 0)
		m_BackgroundPix = c;
	// Output pixel m_Data
	for(i=0;i<n;i++)
		m_Pixels[m_PixelCount++] = c;
	m_CurrentPixelX += n;

	assert(m_CurrentPixelX <= m_SubPicWidth);
	if(m_CurrentPixelX == m_SubPicWidth)
	{
		// End of line...
		// Better code for m_BottomMostPixelT and m_TopMostPixel. Checking the line we 
		// just finished has all equal values.
		int j;
		uchar FirstVal = m_Pixels[m_CurrentPixelY*m_SubPicWidth];
		for(j=0;j<m_SubPicWidth;j++)
		{
			if(m_Pixels[m_CurrentPixelY*m_SubPicWidth+j] != FirstVal)
			{
				// Some of the pixels aren't identical. If m_TopMostPixel hasn't been set
				// yet, now's the time. Also set m_BottomMostPixelT, when we're done with
				// the entire subpic, we'll set m_BottomMostPixel to m_BottomMostPixelT...
				if(m_TopMostPixel == -1) m_TopMostPixel = m_CurrentPixelY;
				m_BottomMostPixelT = m_CurrentPixelY;
				break;
			}
		}
		m_CurrentPixelX = 0;
		m_CurrentPixelY += 2;
		m_PixelCount += m_SubPicWidth;

		// If at the end of a line, we're not at a byte boundary, 4 fill bits of 0 are added.
		assert(m_CurrentBit == 7 || m_CurrentBit == 3);
		if(m_CurrentBit == 3)
		{
			m_CurrentBit = 7;
			m_CurrentByte ++;
		}
		if(!m_SubPicDecoded)
			m_StartOfLine[m_CurrentPixelY] = m_CurrentByte;
	}
	if(m_CurrentPixelY >= m_SubPicHeight)
		return -1;
	else 
		return 0;
}

// RLE decoding... See http://www.mpucoder.com/DVD/spu.html
void CSubPic::RLEDecode(int Field)
{
	int n;
	int c;
	int ret = 0;
	m_CurrentBit = 7;
	if(Field == 0)
	{
		m_PixelCount = 0;
		m_CurrentPixelX = m_CurrentPixelY = 0;
		m_CurrentByte = m_TopFieldOffset;
		m_EndOfField = m_BotFieldOffset;
	}
	else
	{
		m_PixelCount = m_SubPicWidth;
		m_CurrentPixelX = 0;
		m_CurrentPixelY = 1;
		m_CurrentByte = m_BotFieldOffset;
		m_EndOfField = m_OffsetToCmd;
	}

	while(ret == 0)
	{
		n = pop2();
		if(n != 0)
		{
			// 4 bits: nncc
			c = pop2();
			ret = Out(n,c);
		}
		else
		{
			// 00...
			n=pop2();
			if(n != 0)
			{
				// 8 bits: 00nnnncc
				n = 4*n+pop2();
				c = pop2();
				ret = Out(n,c);
			}
			else
			{
				// 0000...
				n=pop2();
				if(n != 0)
				{
					// 12 bits: 0000nnnnnncc
					n = n*16+pop4();
					assert(n>0);
					c = pop2();
					ret = Out(n,c);
				}
				else
				{
					// 16 bits: 000000nnnnnnnncc
					n = pop8();
					c = pop2();
					ret = Out(n,c);
				}
			}
		}
	}
}


int CSubPic::ReadThisPack(uchar* Data, FILE* fd, int LBA, int NavPackLBA, ushort VOBID, 
						  uchar CELLID, uchar StreamID)
{
	assert(m_SPUReadOK == 0);
	m_FilesFrom[m_NumPackInSPU] = fd;
	m_LBAFrom[m_NumPackInSPU] = LBA;

	if(m_DataBytesRead == 0)
	{
		// Find start of subpic data... It's at 0x18+len...I also sure this has a PTS value!
		// That's the only reliable way (?) to make sure this is truly the start of a SPU.
		uint len = (uint) (Data[0x16]);
		assert(len > 0 && (Data[0x15] & 0x80));
		m_SubPicSize = GetShort(Data,0x18+len);
		m_Data = new uchar[m_SubPicSize];

		int ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesRead);
		memcpy(m_Data+m_DataBytesRead,Data+0x18+len,ToRead);
		m_DataBytesRead += ToRead;
		// Now that we have the beginning of the data buffer, we can get the length and the
		// offset to commands.
		m_SubPicSize = GetShort(m_Data,0);
		m_OffsetToCmd = GetShort(m_Data,2);
		assert(m_OffsetToCmd < m_SubPicSize);
		if(m_OffsetToCmd >= m_SubPicSize)
		{
			ERROR2("Inconsistent data in subpic at LBA %d. Offset to commands past end of SPU",m_LBAFrom[0]);
			m_OffsetToCmd = -1;
		}
		m_NavPackLBA = NavPackLBA;
		m_VOBID = VOBID;
		m_CELLID = CELLID;
		m_StreamID = StreamID;
	}
	else
	{
		assert(m_SubPicSize != 0 && m_DataBytesRead != 0);
		// Check that the PES header data length is 0 or that no PTS is present (that would indicate
		// we missed the first SPU pack!).
		uint len = (uint) (Data[0x16]);
		assert(len == 0 || (Data[0x15] & 0x80)==0);
		// Read as much as you can, except if we have enough data.
		int ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesRead);
		memcpy(m_Data+m_DataBytesRead,Data+0x18+len,ToRead);
		m_DataBytesRead += ToRead;
	}
	m_NumPackInSPU++;
	if(m_SubPicSize == m_DataBytesRead)
	{
		// We're done! Now, let's read a few things.
		m_SPUReadOK = true;
		ParseCommands();
		return 1;
	}
	else return 0;
}

int CSubPic::SaveSPU()
{
	if(!m_SPUReadOK)
		return -1;
	int i;
	uchar Buffer[SECTOR_SIZE];
	m_DataBytesWritten = 0;
	for(i=0;i<m_NumPackInSPU;i++)
	{
		// Retrieve the fd and the LBA for the subpic pack we're about to save.
		FILE* fd = m_FilesFrom[i];
		int LBA = m_LBAFrom[i];
		// Seek there and read the sector...
		if(fseek(fd,LBA*SECTOR_SIZE,SEEK_SET) != 0 || fread(Buffer,1,SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to seek/read %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		// Find start of subpic data then save back our subpic data...
		uint len = (uint) (Buffer[0x16]);
		int ToWrite = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesWritten);
		memcpy(Buffer+0x18+len,m_Data+m_DataBytesWritten,ToWrite);
		if(fseek(fd,LBA*SECTOR_SIZE,SEEK_SET) != 0 ||
			fwrite(Buffer,sizeof(char),SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to write %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		m_DataBytesWritten += ToWrite;
	}
	m_Modified = 0;
	return 0;
}

int CSubPic::ReloadSPU()
{
	if(!m_SPUReadOK)
		return -1;
	int i;
	uchar Buffer[SECTOR_SIZE];
	m_DataBytesRead = 0;
	m_SubPicDecoded = 0;
	for(i=0;i<m_NumPackInSPU;i++)
	{
		// Retrieve the fd and the LBA for the subpic pack we're about to reload.
		FILE* fd = m_FilesFrom[i];
		int LBA = m_LBAFrom[i];
		// Seek there and read the sector...
		if(fseek(fd,LBA*SECTOR_SIZE,SEEK_SET) != 0 || fread(Buffer,1,SECTOR_SIZE,fd) != SECTOR_SIZE)
		{
			ERROR3("Failed to seek/read %d-th subpic pack at LBA %d\n",i,LBA);
			return -1;
		}
		// Find start of subpic data then copy the subpic data...
		uint len = (uint) (Buffer[0x16]);
		int ToRead = min(SECTOR_SIZE-0x18-len, m_SubPicSize - m_DataBytesRead);
		memcpy(m_Data+m_DataBytesRead,Buffer+0x18+len,ToRead);
		m_DataBytesRead += ToRead;
	}
	if(m_DataBytesRead != m_SubPicSize)
	{
		ERROR1("Failed to re-load enough data");
		return -1;
	}
	ParseCommands();
	m_Modified = 0;
	m_LastVerticalMove = 0;
	m_VerticalBorders = 0;
	return 0;
}

int CSubPic::DecodeSubpic()
{
	if(m_SPUReadOK == false)
		return -1;
	if(!m_SubPicDecoded)
	{
		// Only set these the first time around. This way we'll keep a memory of them.
		m_StartOfLine[0] = m_TopFieldOffset;
		m_StartOfLine[1] = m_BotFieldOffset;
	}
	RLEDecode(0);
	RLEDecode(1);
	m_SubPicDecoded = true;
	if(m_BottomMostPixel == -1) 
		m_BottomMostPixel = m_BottomMostPixelT;
	// It's possible that the subpic is entirely empty!
	if(m_TopMostPixel == -1) 
		m_TopMostPixel = 0;
	return 0;
}

int CSubPic::OutputPicToTxtFile(const char* FileName)
{
	// Output "bitmap?"
	if(m_SPUReadOK == false)
		return -1;
	FILE* foo = fopen("FileName","w");
	if(!foo)
		return -1;
	int i,j;
	uchar* Ptr = m_Pixels;
	for(i=0;i<m_SubPicHeight;i++)
	{
		for(j=0;j<m_SubPicWidth;j++)
			fprintf(foo,"%d ",*Ptr++);
		fprintf(foo,"\n");
	}
	fclose(foo);
	return 0;
}

int CSubPic::FindCommand(CommandID Command, int DCSQT)
{
	if(m_SPUReadOK == false)
		return -1;
	// Go to DCSQT
	assert(DCSQT < m_DCSQT.size());
	if(DCSQT >= m_DCSQT.size())
		return -1;

	SP_DCSQT* TheDCSQT = &m_DCSQT[DCSQT];
	int i;
	for(i=0;i<TheDCSQT->m_Commands.size();i++)
	{
		if(TheDCSQT->m_Commands[i].m_CommandID == Command)
			return TheDCSQT->m_Commands[i].m_Offset;
	}
	return -1;
}

int CSubPic::ChangeCommand(CommandID OldCommand, CommandID NewCommand, int DCSQT)
{
	int Offset = FindCommand(OldCommand,DCSQT);
	if(Offset == -1)
		return -1;
	// Now change the command!
	m_Data[Offset] = (uchar)NewCommand;
	ParseCommands();
	if(NewCommand != OldCommand)
		m_Modified = true;
	return 0;
}


int CSubPic::ParseCommands(int NumDCSQT, int OffsetToDCSQT)
{
	if(m_SPUReadOK == false)
		return -1;
	if(NumDCSQT == 0)
		m_DCSQT.clear();
	SP_DCSQT Foo;
	m_DCSQT.push_back(Foo);
	SP_DCSQT* TheDCSQT = &m_DCSQT[NumDCSQT];
	// Get time info and offset info.
	TheDCSQT->m_OffsetToThis = (NumDCSQT == 0) ? m_OffsetToCmd : OffsetToDCSQT;
	uchar* Ptr = m_Data + TheDCSQT->m_OffsetToThis;
	TheDCSQT->m_Delay = GetShort(Ptr,0);
	TheDCSQT->m_OffsetToNext = GetShort(Ptr,2);
	TheDCSQT->m_Commands.clear();

	// Now parse commands...
	Ptr += 4;
	bool Done = 0;
	while(!Done)
	{
		// Get command:
		assert(*Ptr != CHG_COLCON);
		// Add a command.
		Commands Temp;
		TheDCSQT->m_Commands.push_back(Temp);
		Commands *TheCmd = &TheDCSQT->m_Commands.back();
		TheCmd->m_CommandID = (CommandID) *Ptr;
		TheCmd->m_Offset = Ptr-m_Data;
		TheCmd->m_NumParams = 0;
		switch(*Ptr)
		{
		case FSTA_DSP : Ptr ++;  break;
		case STA_DSP : Ptr ++; break;
		case STP_DSP : Ptr ++; break;
		case SET_COLOR : 
			TheCmd->m_NumParams = 4;
			TheCmd->m_Params[0] = m_e2Color = m_Colors[3] = Ptr[1]>>4;
			TheCmd->m_Params[1] = m_e1Color = m_Colors[2] = Ptr[1]&0x0F;
			TheCmd->m_Params[2] = m_pColor = m_Colors[1] = Ptr[2]>>4;
			TheCmd->m_Params[3] = m_bColor = m_Colors[0] = Ptr[2]&0x0F;
			Ptr += 3; break;
		case SET_CONTR : 
			TheCmd->m_NumParams = 4;
			TheCmd->m_Params[0] = m_e2Alpha = Ptr[1]>>4;
			TheCmd->m_Params[1] = m_e1Alpha = Ptr[1]&0x0F;
			TheCmd->m_Params[2] = m_pAlpha = Ptr[2]>>4;
			TheCmd->m_Params[3] = m_bAlpha = Ptr[2]&0x0F;
			m_Alphas[3] = 16*m_e2Alpha;
			m_Alphas[2] = 16*m_e1Alpha;
			m_Alphas[1] = 16*m_pAlpha;
			m_Alphas[0] = 16*m_bAlpha;
			Ptr += 3; break;
		case SET_DAREA : 
			TheCmd->m_NumParams = 4;
			TheCmd->m_Params[0] = m_StartX = ((Ptr[1])<<4)|((Ptr[2])>>4);
			TheCmd->m_Params[1] = m_EndX = ((Ptr[2] &0x0F) <<8)|(Ptr[3]);
			TheCmd->m_Params[2] = m_StartY = ((Ptr[4])<<4)|((Ptr[5])>>4);
			TheCmd->m_Params[3] = m_EndY = ((Ptr[5] &0x0F) <<8)|(Ptr[6]);
			if(!m_SubPicDecoded)
			{ m_StartYOrig = m_StartY ; m_EndYOrig = m_EndY ; }
			assert(m_StartX < 1024 && m_EndX < 1024 && m_StartY < 1024 && m_EndY < 1024);
			m_SubPicWidth = m_EndX - m_StartX + 1;
			m_SubPicHeight = m_EndY - m_StartY + 1;
			m_ModSubPicWidth = ((3+m_SubPicWidth)/4)*4;
			Ptr += 7; break;
		case SET_DSPXA : 
			TheCmd->m_NumParams = 2;
			TheCmd->m_Params[0] = m_TopFieldOffset = GetShort(Ptr,1);
			TheCmd->m_Params[1] = m_BotFieldOffset = GetShort(Ptr,3);
			assert(m_TopFieldOffset < 2*m_SubPicSize / 3 && m_BotFieldOffset < 2*m_SubPicSize / 3);
			Ptr += 5; break;
		case CHG_COLCON : 	Done = 1; break;
		case CMD_END : 	Done = 1; break;
		default : 
			// This should NOT happen! Something went wrong here!
			ERROR3("Inconsistent data in subpic at LBA %d. Unknown command %d",m_LBAFrom[0],*Ptr);
			// Setting this to ensure we're exiting!
			TheDCSQT->m_OffsetToNext = TheDCSQT->m_OffsetToThis;
			Done = 1; 
			break;
		}
	}
	if(TheDCSQT->m_OffsetToNext != TheDCSQT->m_OffsetToThis)
		return ParseCommands(m_DCSQT.size(), TheDCSQT->m_OffsetToNext);
	return 0;
}

uchar Colors[3*4] = {0,0,0,255,255,255,0,255,0,0,0,255,};

int CSubPic::SavePicAsPPM(const char* FileName, uchar *CLUT, float FactX, float FactY)
{
	if(m_SPUReadOK == false)
		return -1;
	int NewHeight;
	int NewWidth;
	int TotalByteCount;
	int i,j;

	if(CLUT == 0)
		CLUT = (uchar*)Colors;

	FILE* fd = fopen(FileName,"wb");
	if(fd == 0)
		return -1;
	
	uchar* BitMap = new uchar[(int)(m_SubPicHeight/FactY+1)*(int)(m_SubPicWidth/ FactX+1) * 3];
	uchar* dest = BitMap;

	NewWidth = (int)(m_SubPicWidth/FactX);

	for(i=NewHeight=0,TotalByteCount=0;;i++)
	{
		// Find start of new line. PPM files are stored top to bottom.
		int yindex = (int)(i * FactY + .5f);
		if(yindex >= m_SubPicHeight)
			break;
		uchar* src = m_Pixels + yindex * m_SubPicWidth;
		// Copy this line.
		for(j=0;j<NewWidth;j++)
		{
			// Find pixel location in line.
			int xindex = (int)(j * FactX + .5f);
			// Copy the three RGB components for that pixel. Note the order is different than for BMP files.
			uchar Pick = src[xindex];
			assert(Pick < 16);
			*dest++ = CLUT[3*Pick];
			*dest++ = CLUT[3*Pick+1];
			*dest++ = CLUT[3*Pick+2];
			TotalByteCount+=3;
		}
		NewHeight++;
	}

	fprintf (fd,"P6\n%d %d\n255\n", NewWidth, NewHeight);
	fwrite(BitMap,sizeof(char),TotalByteCount,fd);
	fclose(fd);
	delete[] BitMap;
	return 0;
}

uchar* CSubPic::SaveToBitmap(uchar* Bitmap, long &Width, long &Height, uchar *CLUT)
{
	uchar* dest = Bitmap;
	static uchar  ColorMap[12];

	if(CLUT)
	{
		memcpy(ColorMap,CLUT+3*m_Colors[0],3*sizeof(uchar));
		memcpy(ColorMap+3,CLUT+3*m_Colors[1],3*sizeof(uchar));
		memcpy(ColorMap+6,CLUT+3*m_Colors[2],3*sizeof(uchar));
		memcpy(ColorMap+9,CLUT+3*m_Colors[3],3*sizeof(uchar));
	}
	else
		memcpy(ColorMap,Colors,12*sizeof(char));

	int i,j;
	for(i=m_SubPicHeight-1;i>=0;i--)
	{
		// Find start of new line. PPM files are stored top to bottom.
		uchar* src = m_Pixels + i * m_SubPicWidth;
		// Copy this line.
		for(j=0;j<min(m_SubPicWidth,m_ModSubPicWidth);j++)
		{
			// Copy the three RGB components for that pixel. The order is BGR!
			uchar Pick = 3*src[j];
			assert(Pick < 12);
			*dest++ = ColorMap[Pick+2];	// B
			*dest++ = ColorMap[Pick+1];	// G
			*dest++ = ColorMap[Pick];	// R
		}
		// If the width isn't a multiple of 4, add black pixels.
		for(;j<m_ModSubPicWidth;j++)
		{
			*dest++ = 0;
			*dest++ = 0;
			*dest++ = 0;
		}
	}
	Width = m_ModSubPicWidth;
	Height = m_SubPicHeight;
	return ColorMap;
}

uchar* CSubPic::SaveToAlphaBitmap(uchar* Bitmap, uchar *CLUT)
{
	uchar* dest = Bitmap;
	static uchar  ColorMap[12];
	uchar AlphaMap[4] = {255,255,255,255};


	if(CLUT)
	{
		memcpy(ColorMap,CLUT+3*m_Colors[0],3*sizeof(uchar));
		memcpy(ColorMap+3,CLUT+3*m_Colors[1],3*sizeof(uchar));
		memcpy(ColorMap+6,CLUT+3*m_Colors[2],3*sizeof(uchar));
		memcpy(ColorMap+9,CLUT+3*m_Colors[3],3*sizeof(uchar));
		memcpy(AlphaMap,m_Alphas,4*sizeof(uchar));
	}
	else
	{
		memcpy(ColorMap,Colors,12*sizeof(char));
		memcpy(AlphaMap,m_Alphas,4*sizeof(uchar));
	}

	int i,j;
	for(i=m_SubPicHeight-1;i>=0;i--)
	{
		// Find start of new line. PPM files are stored top to bottom.
		uchar* src = m_Pixels + i * m_SubPicWidth;
		// Copy this line.
		for(j=0;j<min(m_SubPicWidth,m_ModSubPicWidth);j++)
		{
			// Copy the three RGB components for that pixel. The order is BGR!
			uchar Pick = 3*src[j];
			assert(Pick < 12);
			ushort Alpha = AlphaMap[src[j]];
			*dest++ = ColorMap[Pick+2]*Alpha/255;	// B
			*dest++ = ColorMap[Pick+1]*Alpha/255;	// G
			*dest++ = ColorMap[Pick]*Alpha/255;		// R
			*dest++ = (uchar)Alpha;					// Alpha
		}
		// If the width isn't a multiple of 4, add transparent pixels.
		for(;j<m_ModSubPicWidth;j++)
		{
			*dest++ = 0;
			*dest++ = 0;
			*dest++ = 0;
			*dest++ = 0;
		}
	}
	return ColorMap;
}


void CSubPic::WriteDAREA(ushort StartX, ushort StartY, ushort EndX, ushort EndY)
{
	int Offset = FindCommand(SET_DAREA);
	if(Offset != -1)
	{
		m_Data[Offset+1] = (uchar)(StartX >> 4);
		m_Data[Offset+2] = (uchar)((StartX & 0x000F)<<4);
		m_Data[Offset+2] |= (uchar)(EndX >> 8);
		m_Data[Offset+3] = (uchar)(EndX & 0x00FF);
		m_Data[Offset+4] = (uchar)(StartY >> 4);
		m_Data[Offset+5] = (uchar)((StartY & 0x000F)<<4);
		m_Data[Offset+5] |= (uchar)(EndY >> 8);
		m_Data[Offset+6] = (uchar)(EndY & 0x00FF);
	}
}

void CSubPic::WriteDSPXA(ushort TopOffset, ushort BotOffset)
{
	int Offset = FindCommand(SET_DSPXA);
	if(Offset != -1)
	{
		PutShort(m_Data,Offset+1,TopOffset);
		PutShort(m_Data,Offset+3,BotOffset);
	}
}

int CSubPic::MoveSubPic(int NumPixels)
{
	// Decode subpic if needed.
	if(!m_SubPicDecoded)
	{
		if(DecodeSubpic() == -1)
			return -1;
	}
	// First of, "reset" the subpic, i.e. undo any move that might have been done before...
	m_TopFieldOffset = m_StartOfLine[0];
	m_BotFieldOffset = m_StartOfLine[1];
	m_StartY = m_StartYOrig;
	m_EndY = m_EndYOrig;
	m_SubPicHeight = m_EndY - m_StartY + 1;
	// Make sure NumPixels is even!
	NumPixels = min(NumPixels,m_StartY-m_VerticalBorders+m_TopMostPixel);
	NumPixels = max(m_StartY+m_BottomMostPixel-m_DisplayHeight+m_VerticalBorders,NumPixels);
	NumPixels = 2*(NumPixels/2);

	m_LastVerticalMove = NumPixels;
//	if(NumPixels >= m_SubPicHeight || -NumPixels >= m_SubPicHeight)
//		return -1;

	if(NumPixels > 0)
	{
		// Moving subpic up.
		// This is tricky. If m_StartY isn't at the top of the picture (say 1, for example),
		// then we can move the subpic simply by adjusting m_StartY and m_EndY without changing
		// the pointers. We can move that way by NumPix1. To move further, we have to change
		// the pointers. We move that way by NumPix2.
		int NumPix1 = 2*((m_StartY - 1)/2);	// Must be even!
		NumPix1 = min(NumPix1,NumPixels);
		int NumPix2 = NumPixels-NumPix1;
		NumPix2 = min(NumPix2,m_SubPicHeight-3);	// so m_EndY >= m_StartY+2!
		assert(NumPix2+1 < m_SubPicHeight);
		// Moving subpics up. Set the offset to the top and bottom fields to skip the top
		// lines, then adjust the endy value...
		// First, find offset to NumPixels line.
		assert(NumPix2 >= 0);
		m_TopFieldOffset = m_StartOfLine[NumPix2];
		m_BotFieldOffset = m_StartOfLine[NumPix2+1];
		// Adjust EndY.
		assert(m_EndY > NumPix2+NumPix1);
		m_EndY -= (NumPix2+NumPix1);
		m_StartY -= NumPix1;
		// Then write all this back!
		WriteDSPXA(m_TopFieldOffset,m_BotFieldOffset);
		WriteDAREA(m_StartX, m_StartY, m_EndX, m_EndY);
		m_SubPicHeight = m_EndY - m_StartY + 1;
	}
	else
	{
		// Moving subpics down. Simply adjust the starty value... 
		// Adjust StartY and EndY, the hope is that if EndY is too large, the subpic will simply
		// be cropped.
		m_StartY += (-NumPixels);
		m_EndY += (-NumPixels);
		// Then write all this back!
		WriteDSPXA(m_TopFieldOffset,m_BotFieldOffset);
		WriteDAREA(m_StartX, m_StartY, m_EndX, m_EndY);
		m_SubPicHeight = min(m_EndY,m_DisplayHeight) - m_StartY + 1;
	}
	m_Modified = 1;
	return 0;
}

int CSubPic::MoveSubPicInFrame(int MinTopMostPos)
{
	// Avoid moving subpics that are actually highlights.
	if(m_HasButtons)
		return 0;

	// Bottom margin is the same as the top one.
	int MaxBottomMostPos = m_DisplayHeight - MinTopMostPos;
	m_VerticalBorders = MinTopMostPos;

	// m_StartY does not always indicate the current amount by which the subpics were
	// moved (for example if they were moved up by modifying the pointers).
	// Instead, m_StartYOrig - m_LastVerticalMove tells me the true start Y
	// position, after the last move.
	long CurrentStartY = m_StartYOrig - m_LastVerticalMove;

	// Make sure the subpic is either above our bottom limit or below our top one. If it's violating
	// both limits we can't do anything!
	assert(CurrentStartY+m_TopMostPixel >= MinTopMostPos || 
		CurrentStartY+m_BottomMostPixel <= MaxBottomMostPos);

	int MoveBy = 0;
	// Check whether our subpic is below MinTopMostPos.
	if(CurrentStartY+m_TopMostPixel < MinTopMostPos)
	{
		// We need to move this guy down! MoveBy will be negative.
		MoveBy = CurrentStartY+m_TopMostPixel - MinTopMostPos;
		// Check that by moving this guy down, we're not going past the bottom limit.
		if(CurrentStartY+m_BottomMostPixel-MoveBy > MaxBottomMostPos)
		{
			assert(0);
			return 0;
		}
	}
	else if(CurrentStartY+m_BottomMostPixel > MaxBottomMostPos)
	{
		// We need to move this guys up! MoveBy will be positive
		MoveBy = CurrentStartY+m_BottomMostPixel - MaxBottomMostPos;
		// Check that by moving this guy up, we're not going past the toplimit.
		if(CurrentStartY+m_TopMostPixel-MoveBy < MinTopMostPos)
		{
			assert(0);
			return 0;
		}
	}
	MoveSubPic(m_LastVerticalMove + MoveBy);
	return m_LastVerticalMove + MoveBy;
}


void CSubPic::ChangeTransparency(int Transparency, int PixelSelect)
{
	int Offset = FindCommand(SET_CONTR);
	if(Offset == -1)
		return;
	uchar* Ptr = m_Data+Offset;

	if(PixelSelect < 4)
	{
		m_Alphas[PixelSelect] = Transparency*16;
		switch(PixelSelect)
		{
		case 0 : m_bAlpha = Transparency; 	break;
		case 1 : m_pAlpha = Transparency; 	break;
		case 2 : m_e1Alpha = Transparency;	break;
		case 3 : m_e2Alpha = Transparency;	break;
		}
	}
	else
	{
		// Change the transparency of all pixels, except for the background pixels.
		if(m_BackgroundPix != 0) m_bAlpha = Transparency, m_Alphas[0] = Transparency*16; 
		if(m_BackgroundPix != 1) m_pAlpha = Transparency, m_Alphas[1] = Transparency*16; 
		if(m_BackgroundPix != 2) m_e1Alpha = Transparency, m_Alphas[2] = Transparency*16; 
		if(m_BackgroundPix != 3) m_e2Alpha = Transparency, m_Alphas[3] = Transparency*16; 
	}

	// Put that in our data!
	Ptr[2] &= 0xF0; Ptr[2] |= m_bAlpha;		
	Ptr[2] &= 0x0F; Ptr[2] |= (m_pAlpha << 4);
	Ptr[1] &= 0xF0; Ptr[1] |= m_e1Alpha;		
	Ptr[1] &= 0x0F; Ptr[1] |= (m_e2Alpha << 4);

	m_Modified = true;
}

int CSubPic::GetMeanTransparency()
{
	int Average = 0;
	if(m_BackgroundPix != 0) Average += m_bAlpha;
	if(m_BackgroundPix != 1) Average += m_pAlpha;
	if(m_BackgroundPix != 2) Average += m_e1Alpha;
	if(m_BackgroundPix != 3) Average += m_e2Alpha;
	return (int)((Average+.5f)/3);
}

int CSubPic::GetCommand(char* Command, int &DCSQT, int &CommandIndex)
{
	if(DCSQT >= m_DCSQT.size())
		return -1;
	SP_DCSQT *TheDCSQT = &m_DCSQT[DCSQT];
	if(CommandIndex == -1)
	{
		if(DCSQT == 0)
			sprintf(Command,"DCSQT: %d\n",DCSQT);
		else
			sprintf(Command,"DCSQT: %d -- Delay %.3fs\n",DCSQT,
			(float)TheDCSQT->m_Delay*1024/90000);
		goto end;
	}

	switch(TheDCSQT->m_Commands[CommandIndex].m_CommandID)
	{
	case FSTA_DSP : sprintf(Command,"  Forced Start\n"); break;
	case STA_DSP : sprintf(Command,"  Start\n"); break;
	case STP_DSP : sprintf(Command,"  Stop\n"); break;
	case SET_COLOR : sprintf(Command,"  Set Color: e2=%d e1=%d p=%d b=%d\n",m_e2Color,m_e1Color,m_pColor,m_bColor); break;
	case SET_CONTR : sprintf(Command,"  Set Contrast: e2=%d e1=%d p=%d b=%d\n",m_e2Alpha,m_e1Alpha,m_pAlpha,m_bAlpha); break;
	case SET_DAREA : sprintf(Command,"  Set Display Area: sx=%d ex=%d sy=%d ey=%d \n", m_StartX, m_EndX, m_StartY, m_EndY); break;
	case SET_DSPXA : sprintf(Command,"  Set field indexes\n"); break;
	case CHG_COLCON : sprintf(Command,"  CHG_COLCON\n"); break;
	case CMD_END : sprintf(Command,"  End of commands\n"); break;
	}

end:
	CommandIndex++;
	if(CommandIndex >= TheDCSQT->m_Commands.size())
	{
		CommandIndex = -1;
		if(DCSQT < m_DCSQT.size()-1) DCSQT++;
		else
		{
			DCSQT = 0; return -1;
		}
	}
	return 0;
}

