#ifndef SubPic_H
#define SubPic_H

#include "includes.h"
// STL stuff for our arrays.
#include <vector>

class CCell;
class SP_DCSQT;

// Command enum
typedef enum
{
	FSTA_DSP = 0,
	STA_DSP = 1,
	STP_DSP  = 2,
	SET_COLOR = 3,
	SET_CONTR = 4,
	SET_DAREA = 5,
	SET_DSPXA = 6,
	CHG_COLCON = 7,
	CMD_END = 0xFF,
} CommandID;

// Structure that contains the info about 1 command.
typedef struct
{
	CommandID	m_CommandID;	// Command ID.
	ushort		m_Offset;		// Offset to this command in m_Data.
	short		m_NumParams;	// How many parameters.
	ushort		m_Params[4];	// Command Parameters.
} Commands;
typedef std::vector<Commands> CommandsV;

// Class containing all the info about 1 CDSQT (group of commands)
#define MAX_NUM_COMMANDS 16
class SP_DCSQT
{
public:
	ushort		m_Delay;		// Delay befor execution
	ushort		m_OffsetToThis;	// Offset to start of this CDSQT
	ushort		m_OffsetToNext;	// Offset to next CDSQT. 0 if none
	CommandsV	m_Commands;		// The commands;
};

typedef std::vector<SP_DCSQT> DCSQTVect;


class CSubPic
{
public: 
	CSubPic();
	~CSubPic();

	void	Reset(bool ResetAll=true);

	// We look starting at PackIndex. If PackIndex = -1, we read the next SPU.
	int		WriteSubPacks(FILE* fd);
	// This decodes the subpic image.
	int		DecodeSubpic();
	// This outputs the subpic image as a text file, 1 index into the CLUT per pixel.
	int		OutputPicToTxtFile(const char* FileName);
	// This outputs the subpic image as a ppm file.
	int		SavePicAsPPM(const char* FileName, uchar *CLUT = 0, float FactX=1, float FactY=1);
	// This stores the subpic image as a bitmap.
	uchar*	SaveToBitmap(uchar* Bitmap, long &Width, long &Height, uchar *CLUT = 0);
	uchar*	SaveToAlphaBitmap(uchar* Bitmap, uchar *CLUT = 0);

	// This reads the pack in Data read from file fd at pack LBA (relative to start of file).
	// return -1 if something went wrong, 0 if it's still missing data, 1 when SPU is complete.
	// NavPackLBA is the LBA of the previous navpack, relative to start of VTS (not start of file).
	int		ReadThisPack(uchar* Data, FILE* fd, int LBA, int NavPackLBA, 
		ushort VOBID, uchar CELLID, uchar StreamID);

	// This saves the entire SPU (all its subpic packs) to their original locations.
	int		SaveSPU();

	// This reloads the entire SPU from file.
	int		ReloadSPU();

	// Retrieves the command in a human-readable form! Increments CommandIndex and DCSQT,
	// return -1 if no more...
	int		GetCommand(char* Command, int &DCSQT, int &CommandIndex);

	// This can move the subpics up or down by NumPixels, which should be even. Returns 0 if
	// it could, or the number of pixels it was able to move the subpics by. Positive numbers
	// to move up, negative to move down.
	int		MoveSubPic(int NumPixels);

	// Here, the subpic is moved so the topmost pixel is no higher than MinTopMostPos and the bottommost
	// pixel is more than MinTopMostPos above the bottom of the picture. 
	// This is useful to move subpics inside a box, if needed.
	int		MoveSubPicInFrame(int MinTopMostPos);

	// This changes the transparency of the selected pixel type. Transparency should be between 0 and 15.
	// PixelSelect can be 0,1,2,3 (for b,p,e1,e2) or 4 for all.
	void	ChangeTransparency(int Transparency, int PixelSelect);
	// This returns the average transparency of non-background pixels...
	int		GetMeanTransparency();

	// Find the command Command in the display control sequence table DCSQT. Return the
	// byte offset in m_Data to the Command itself. -1 if not found.
	int		FindCommand(CommandID Command, int DCSQT = 0);

	// This replaces a command by another one. You need to know what you're doing here! 
	int		ChangeCommand(CommandID OldCommand, CommandID NewCommand, int DCSQT = 0);

	int		m_CurrentBit;		// Next bit to be poped
	int		m_CurrentByte;		// Current byte we're popping from!
	int		m_PixelCount;		// Count of output m_pixels.
	int		m_CurrentPixelX;	// x-value for current output pixel.
	int		m_CurrentPixelY;	// y-value for current output pixel.
	int		m_SubPicWidth;		// Line width.
	int		m_ModSubPicWidth;	// For bitmaps, the width has to be a multiple of 4. This is it.
	int		m_SubPicHeight;		// Number of lines.

	ushort	m_SubPicSize;		// The overall byte size of the subpic data, including commands etc.
	ushort	m_OffsetToCmd;		// Byte offset in m_Data to commands.

	ushort	m_StartYOrig;		// Start of subpic bounding rectangle in original subpic.
	ushort	m_EndYOrig;			// End of subpic bounding rectangle in original subpic.
	ushort	m_StartX;			// Start of subpic bounding rectangle.
	ushort	m_StartY;			// Start of subpic bounding rectangle.
	ushort	m_EndX;				// End of subpic bounding rectangle.
	ushort	m_EndY;				// End of subpic bounding rectangle.
	ushort	m_TopFieldOffset;	// Offset to first pix data of top field in m_Data.
	ushort	m_BotFieldOffset;	// Offset to first pix data of bottom field in m_Data.
	uchar	m_Colors[4];		// Colors for e2, e1, p, and b pixels, as defined in commands.
	uchar	m_e2Color;			// Colors for e2, e1, p, and b pixels, as defined in commands.
	uchar	m_e1Color;
	uchar	m_pColor;
	uchar	m_bColor;

	uchar	m_e2Alpha;			// Alpha for e2, e1, p, and b pixels, as defined in commands.
	uchar	m_e1Alpha;
	uchar	m_pAlpha;
	uchar	m_bAlpha;
	uchar	m_Alphas[4];
	int		m_BackgroundPix;	// Which of the four pixel types seems to be the background.

	static	int m_DisplayHeight;// Vertical size of video display.

	int		m_LastVerticalMove;	// Store the amount of vertical moving last applied.
	int		m_VerticalBorders;	// Minimum space between top and bottom of pic, and subpic data.

	uchar	*m_Data;			// Binary data, including pixel data and commands.
static	uchar	*m_Pixels;		// Decoded pixel data (values between 0 and 15, index into KLUT).
	// For each line, m_LastSetPixel indicates the last pixel whose value is set individually, as
	// opposed to "until end of line".
	ushort	*m_LastSetPixel;
	ushort	*m_StartOfLine;		// Index in m_Data of data for first pixel for this line.
	ushort	m_RightMostPixel;	// Max value in m_LastSetPixel;
	short	m_TopMostPixel;		// First line with a pixel that's defined individually.
	short	m_BottomMostPixel;	// Last line with a pixel that's defined individually.
	short	m_BottomMostPixelT;	// A temporary value... 

//	int			m_NumDCSQT;		// Number of command sets (usually 1 or 2).
	DCSQTVect	m_DCSQT;		// Vector of DCSQTs (sets of commands).

	int		m_Selected;			// Indicates this SPU is selected for batch processing.
	int		m_Modified;			// Indicates this SPU's data is different than what's on disk.
	ushort	m_VOBID;			// These indicate the VCID this subpic is in.
	uchar	m_CELLID;			
	uchar	m_StreamID;			// The stream ID for this spu.
	int		m_HasButtons;		// Will be non-zero if the neighboring navpacks indicate buttons.

	int		m_NumPackInSPU;		// How many subpic packs form this SPU.
	FILE*	m_FilesFrom[32];	// Files the various packs that form this SPU come from.
	int		m_LBAFrom[32];		// Where in above file were the subpic packs located.
	int		m_NavPackLBA;		// LBA of preceding nav pack, relative to start of VTS.
	int		m_SPUNumber;		// A number associated with this SPU.

protected:

	uchar	pop2();				// Pops 2 bits from m_Data.
	uchar	pop4();				// Pops 4 bits from m_Data.
	uchar	pop8();				// Pops 8 bits from m_Data.
	int		Out(int n, int c);	// Outputs pixel data into m_Pixels;
	void	RLEDecode(int Field);	// Decodes 1 field of the subpic data.

	int		DoWriteSubPack(FILE* fd);

	// Parse all commands in all DCSQT. Is called recursively.
	int		ParseCommands(int NumDCSQT = 0, int OffsetToDCSQT = 0);

	void	WriteDAREA(ushort StartX, ushort StartY, ushort EndX, ushort EndY);
	void	WriteDSPXA(ushort TopOffset, ushort BotOffset);

	int		m_EndOfField;		// End of data area for the current field being decoded.
	uint	m_DataBytesRead;	// How many data bytes have been read so far.
	uint	m_DataBytesWritten;	// How many data bytes have been written so far.

	bool	m_SPUReadOK;		// Indicates whether Subpic unit was read correctly.
	bool	m_SubPicDecoded;	// Indicates whether Subpic unit was decoded.

};








#endif