#include "SubPicEngine.h"
#include <stdio.h>
#include <objbase.h>
#include "Cell.h"
#include "Utils.h"
#include "SubPic.h"
#include "IFOUtils.h"

typedef unsigned char uchar;

#define VERS "DVDSubEdit -- Beta 0.73 By Jeanl"

#define CHAR_SIZE 256
// This stuff is global and accessed by all the classes, for messages and errors.
// The string where messages are stored.
char Message[CHAR_SIZE];
// Our default function to handle regular and error messages.
void DefaultLogMessage(char* AMessage, int Langage)
{
	printf(AMessage);
	strcpy(Message,AMessage);
#if _DEBUG
	OutputDebugString(Message);
	OutputDebugString("\n");
#endif
}
// Global pointers that classes use to send regular and error messages.
void (*pMessageFunction)(char* AMessage, int Langage) = DefaultLogMessage;
void (*pErrorFunction)(char* AMessage, int Langage) = DefaultLogMessage;
int	 (*FrameSelectFunction)(int VobId, int CellId, int StartLBA, int EndLBA, int Default, 
							int NumLBA, int NormallySkipped) = 0;

// I need this destructor so the CLUT arrays are automatically destroyed.
PGC_CLUT_Info::~PGC_CLUT_Info()
{
	int i;
	for(i=0;i<m_CLUT.size();i++)
		delete[] m_CLUT[i];
}

CSubPicEngine::CSubPicEngine()
{
	m_CLUTs = 0;
	strcpy(m_Version,VERS);
	Reset();
}
	
CSubPicEngine::~CSubPicEngine()
{
	Reset();
	delete[] m_CLUTs;
	int i;
	for(i=0;i<32;i++)
	{
		m_SubPics[i].clear();
	}
}

void CSubPicEngine::ScanFilesForSPUs()
{
	// Scan all the vob files for subpics...
	int i,LBA,ret;
	int TotNumLBA = 0;
	int CumulLBA = 0;
	int NavPackNumButtons = 0;
	DWORD End;

	for(i=0;i<m_NumVOBFiles;i++)
	{
		TotNumLBA += GetFileSizekB(m_FileName[i])/2;
		fseek(m_FileFD[i],0,SEEK_SET);
	}

	CPack ThePack;
	int LastNavPack = -1;
	m_TotNumSPU = 0;
	m_MasterList.clear();
	DWORD Start = timeGetTime();
	m_Phase = Processing_Scanning_VOB;
	float SmoothRate = 0;
	VideoFormat = FORMAT_UNKNOWN;
	for(i=0;i<m_NumVOBFiles;i++)
	{
		for(LBA=0;;LBA++,CumulLBA++)
		{
			if(m_AbortNow)
			{
				m_Phase = Processing_Aborted;
				goto end ;
			}
			m_PercentScanned = 100*(float)CumulLBA / TotNumLBA;
			if((CumulLBA+1) % 2000 == 0)
			{
				DWORD End = timeGetTime();
				float RateMBpS = (1000*(float)CumulLBA*2/1024/(End-Start));
				if(SmoothRate == 0) SmoothRate = RateMBpS;
				else SmoothRate = .95f*SmoothRate + .05f*RateMBpS;
				m_RemainingTimeSec = (TotNumLBA - CumulLBA)*2/1024 / SmoothRate;
				MESSAGE5("%.2f MB scanned (%.2f MB/s) -- %d subpics -- %.0f secs to go",(float)CumulLBA/2/1024,
					RateMBpS,m_TotNumSPU, m_RemainingTimeSec);
				
			}
			ret=(fread(m_Buffer,1,SECTOR_SIZE,m_FileFD[i]) == SECTOR_SIZE);
			if(ret != 1)
				break;
			ushort VOBID,vobid;
			uchar CELLID,cellid;
			ThePack.FindPackType(m_Buffer, vobid, cellid);
			if(ThePack.m_PackType != PackTypeSupPic)
			{
				if(ThePack.m_PackType == PackTypeNavPack)
				{
					VOBID = vobid; CELLID = cellid;
					LastNavPack = CumulLBA;
					NavPackNumButtons = ThePack.m_NumButtons;
				}
				else if(ThePack.m_PackType == PackTypeVideo && VideoFormat == FORMAT_UNKNOWN)
				{
					ThePack.FindFrameType(m_Buffer,0,&m_AspectRatio);
					if(VideoFormat != FORMAT_UNKNOWN)
					{
						CSubPic::m_DisplayHeight = (VideoFormat == FORMAT_PAL) ? 576 : 480;
						JustFoundVideoFormat();
					}
				}
				continue;
			}
			uchar StreamID = ThePack.m_StreamID-0x20;
			CSubPic* Temp = m_SubPics[StreamID][m_NumSubPics[StreamID]];

			m_TopStreamID = max(m_TopStreamID,(int)StreamID);
			if(Temp == NULL)
			{
				Temp = new CSubPic;
				m_SubPics[StreamID][m_NumSubPics[StreamID]] = Temp;
			}
			Temp->m_HasButtons = max(NavPackNumButtons,Temp->m_HasButtons);
			if(Temp->ReadThisPack(m_Buffer, m_FileFD[i], LBA, LastNavPack, VOBID, CELLID,StreamID) == 1)
			{
				if(Temp->m_NavPackLBA == -1)
				{
					// This can happen if there hasn't been a navpack before this subpic, for example
					// because we started scanning in the middle of a VTS. At this point, I'm simply
					// going to discard this guy!
					delete Temp;
					Temp = m_SubPics[StreamID][m_NumSubPics[StreamID]] = 0;
				}
				else
				{
					if(m_NumSubPics[StreamID] == 0)
						JustFoundStream(StreamID);
					if(m_SelectedSubPic == 0)
						m_SelectedSubPic = Temp;
					m_NumSubPics[StreamID]++;
					m_SubPics[StreamID].push_back(0);
					Temp->m_SPUNumber = m_TotNumSPU;
					m_MasterList.push_back(Temp);
					m_TotNumSPU++;
					if(m_VCIDLength == 0 || 256*(uint)VOBID+CELLID != m_VCIDList[m_VCIDLength-1])
					{
						m_VCIDList[m_VCIDLength++] = 256*(uint)VOBID+CELLID;
						JustFoundVCID(256*(uint)VOBID+CELLID);
					}
					JustAddedSPU(StreamID,m_NumSubPics[StreamID]-1);
				}
			}
		}
	}
	End = timeGetTime();
	MESSAGE4("%.2f MB scanned (%.2f MB/s) -- %d subpics",(float)CumulLBA/2/1024,
		1000*(float)CumulLBA*2/1024/(End-Start),m_TotNumSPU);
	m_Phase = Processing_Done;

end:
	ScanIFOs();

//CSubPic* Temp = m_SubPics[0][1];
//Temp->ReloadSPU();
}

void CSubPicEngine::ScanIFOs()
{
	int i,j, NumFound;
	// Scan IFOs, looking for each VCID, and getting the CLUT for the PGCs which reference
	// them...
	// First, get IFO file.
	ushort *LU;
	ushort ALU[1024];
	CIFOLoader IFO(GetIFO(m_FileName[0]));
	char Num = m_FileName[0][strlen(m_FileName[0])-5];
	if(Num > '0' && Num <= '9')
		LU = 0;
	else
		LU = ALU;
//	assert(IFO.m_LoadedOK == true);
	if(!IFO.m_LoadedOK)
		return;

	// Allocate CLUTs.

	delete[] m_CLUTs; m_CLUTs = 0;
	m_CLUTs = new PGC_CLUT_Info[m_VCIDLength];

	// Now go through all the VCID, and find which PGCs referenced it..
	ushort PGC[1024]; 
	ushort CELLOut[1024];
	for(i=0;i<m_VCIDLength;i++)
	{
		NumFound = IFO.FindThisCell(m_VCIDList[i] / 256, m_VCIDList[i] % 256, LU, PGC, CELLOut);
		m_CLUTs[i].m_NumCLUT = NumFound;
		// For each PGC, store the clut in our table...
		for(j=0;j<NumFound;j++)
		{
			m_CLUTs[i].m_PGC.push_back(PGC[j]);
			m_CLUTs[i].m_LU.push_back((LU)?LU[j]:0);
			uchar *Foo  = new uchar[3*16];
			m_CLUTs[i].m_CLUT.push_back(Foo);
			IFO.GetCLUT(m_CLUTs[i].m_CLUT[j],(LU)?LU[j]:0,PGC[j],true);
		}
		// Get the language for the first PGC only...
		if(j>0)
		{
			int m;
			for(m=0;m<32;m++)
			{
				IFO.GetPgcSubpicLanguage(m_CLUTs[i].m_Languages+2*m,(LU)?LU[0]:0,PGC[0],0x20+m);
			}
		}
	}
	m_CLUTLength = m_VCIDLength;
	IFO.CloseIFO();
}

void CSubPicEngine::ReloadAllSPUs()
{
	int i,j;
	for(i=0;i<32;i++)
	{
		for(j=0;j<m_NumSubPics[i];j++)
		{
			CSubPic* Temp = m_SubPics[i][j];
			Temp->ReloadSPU();
		}
	}
}

void CSubPicEngine::SaveAllModifiedSPUs()
{
	int i,j;
	for(i=0;i<32;i++)
	{
		for(j=0;j<m_NumSubPics[i];j++)
		{
			CSubPic* Temp = m_SubPics[i][j];
			if(Temp->m_Modified)
				Temp->SaveSPU();
		}
	}
}

void CSubPicEngine::Reset()
{
	int i;
	m_AbortNow = 0;
	m_TotNumSPU = 0;
	m_PruneLength = 0;
	m_PrunedList.clear();
	m_PercentScanned = 0;
	m_Phase = Processing_Idle;
	m_NumVOBFiles = 0;
	m_TopStreamID = 0;
	m_SelectedSubPic = 0;

	m_VCIDLength = 0;
	m_CLUTLength = 0;
	m_AspectRatio = 0;
	memset(m_NumSubPics,0,32*sizeof(int));
	for(i=0;i<32;i++)
	{
		int j;
		for(j=0;j<m_SubPics[i].size();j++)
			delete m_SubPics[i][j];
		m_SubPics[i].clear();
		m_SubPics[i].push_back(0);
	}
	m_MasterList.clear();
	
	for(i=0;i<10;i++)
	{
		m_FileFD[i] = 0;
	}
}	

int CSubPicEngine::AddFile(const char* FileName)
{
	strcpy(m_FileName[m_NumVOBFiles],FileName);
	m_FileFD[m_NumVOBFiles] = fopen(FileName,"rb+");
	if(m_FileFD[m_NumVOBFiles] == 0)
		return -1;
	m_NumVOBFiles++;
	return 0;
}

int CSubPicEngine::PruneSPUList(int CurrentSelected, int StreamID, int StartLBA, int EndLBA, ushort VOBID, uchar CELLID, int Modified)
{
	int i;
	int N;
	int NavPackLBA = -1;
	int PrevSPUNumber = -1;
	int RetIndex = -1;
	CSubPicPV Ptr;

	if(CurrentSelected < m_PruneLength)
	{
		NavPackLBA = m_PrunedList[CurrentSelected]->m_NavPackLBA;
		PrevSPUNumber = m_PrunedList[CurrentSelected]->m_SPUNumber;
	}
	m_PruneLength = 0;
	m_PrunedList.clear();
	if(StreamID >= 0)
	{
		// A streamID was selected. Select the pointer that corresponds to that stream ID.
		if(StreamID > m_TopStreamID)
			goto end;
		Ptr = m_SubPics[StreamID];
		N = m_NumSubPics[StreamID];
	}
	else
	{
		// Otherwise, select the master pointer to get to all the SPUs.
		Ptr = m_MasterList;
		N = m_TotNumSPU;
	}
	for(i=0;i<N;i++)
	{
		CSubPic* Temp = Ptr[i];
		if(StartLBA != -1 && Temp->m_LBAFrom[0] < StartLBA)
			continue;
		if(EndLBA != -1 && Temp->m_LBAFrom[0] > EndLBA)
			continue;
		if(VOBID != 0 && Temp->m_VOBID != VOBID)
			continue;
		if(CELLID != 0 && Temp->m_CELLID != CELLID)
			continue;
		if(Modified != 0 && Temp->m_Modified == 0)
			continue;
		assert(Ptr[i] != 0);
		// The new list might not have the previously selected SPU. Let's find one that corresponds to
		// the same image (navpack).
		if(RetIndex == -1 && NavPackLBA <= Ptr[i]->m_NavPackLBA)
			RetIndex = m_PruneLength;
		// But if the previously selected SPU is in the new list, select this guy!
		if(PrevSPUNumber == Ptr[i]->m_SPUNumber)
			RetIndex = m_PruneLength;
		m_PrunedList.push_back(Ptr[i]);m_PruneLength++;
	}
end: 
	return RetIndex;
}

int CSubPicEngine::RetreiveNextCommand(char* Command)
{
	if(m_SelectedSubPic == 0)
		return -1;
	static int CurrentDCSQT = 0;
	static int CurrentCommand = -1;
	return m_SelectedSubPic->GetCommand(Command, CurrentDCSQT, CurrentCommand);
}

