#ifndef VTS_CLEAN_H
#define VTS_CLEAN_H

#include "includes.h"
#include <vector>

class CSubPic;

enum ProcessPhase
{
	Processing_Idle,
	Processing_Scanning_VOB,
	Processing_Done,
	Processing_Aborted,
	Processing_Applying,
	Processing_Reverting,
	Processing_Saving,
};

typedef std::vector<int> intV;
typedef std::vector<unsigned char*> char16V;

class PGC_CLUT_Info
{
public:
	~PGC_CLUT_Info();
	int		m_NumCLUT;
	intV	m_LU;
	intV	m_PGC;
	char16V	m_CLUT;
	char	m_Languages[64];
};

typedef std::vector<CSubPic*> CSubPicPV;

class CSubPicEngine
{
public:
	CSubPicEngine();
	~CSubPicEngine();

	// Call this before you add any file via AddFile() for later scanning.
	virtual void	Reset();
	// Call this repeatedly to add files for later scanning.
	int				AddFile(const char* FileName);
	// After adding all the files via AddFile(), call this to start scanning.
	virtual void	ScanFilesForSPUs();
	// This scans the IFOs. It's called automatically in ScanFilesForSPUs();
	virtual void	ScanIFOs();
	// Call this to reload all the SPU from file. This does NOT rescan for subpic packs, instead
	// relies on the previous scan.
	virtual void	ReloadAllSPUs();
	// Call this to save all modified SPUs back to disk.
	virtual void	SaveAllModifiedSPUs();

	// You can call this repeatedly to retrieve command info in a human-readable way. This applies
	// to m_SelectedSubPic.
	int				RetreiveNextCommand(char* Command);

	// This will be called each time an SPU is added during the scan. You can override it and do what
	// you will with it!
	virtual void	JustAddedSPU(int StreamIndex, int SPUIndex) {return;}
	virtual void	JustFoundStream(int StreamIndex) {return;}
	virtual void	JustFoundVCID(uint VCID) {return;}
	virtual void	JustFoundVideoFormat(){return;}

	// This creates m_PrunedList based on their stream ID, LBAs, or VCID.
	virtual int		PruneSPUList(int CurrentSelected, int StreamID = -1, int StartLBA = -1, int EndLBA = -1,
		ushort VOBID = 0, uchar CELLID = 0, int Modified = 0);

	ProcessPhase	m_Phase;					// What are we currently doing?

	int				m_NumVOBFiles;				// How many files we have to scan.
	char			m_FileName[10][1024];		// Names of files to be scanned for SPUs.
	FILE*			m_FileFD[10];				// File Descr. of files to be scanned for SPUs.
	bool			m_AbortNow;					// Set this to true to stop scanning.
	uchar			m_Buffer[SECTOR_SIZE];		// Temporary storage to read sectors.


	float			m_PercentScanned;			// How far along in percent have we scanned.
	float			m_RemainingTimeSec;			// How long till we're done!

	CSubPicPV		m_SubPics[32];				// Two-dimensional array of pointers to SPU objects.
//												// m_SubPics[10][100] is the 100th SPU in stream 32+10.
	CSubPicPV		m_MasterList;				// List of all SPUs, in the order they were found.
	CSubPicPV		m_PrunedList;				// List of SPUs pruned according to filters.
	int				m_TotNumSPU;				// Overall number of found SPUs.
	int				m_PruneLength;				// Length of pruned list.
	uint			m_VCIDList[1024];			// List of found VCIDs (VOBID*128+CELLID).
	int				m_VCIDLength;
	PGC_CLUT_Info	*m_CLUTs;					// Array of CLUTS info
	int				m_CLUTLength;				// Length of this array. Is m_VCIDLength in the end.

	int				m_NumSubPics[32];			// How many SPUs per stream.
	int				m_TopStreamID;				// Highest stream ID for which we found an SPU.

	CSubPic			*m_SelectedSubPic;			// Subpic currently selected for display/modif.
	char			m_Version[512];

	uchar			m_AspectRatio;				// Aspect ratio. 1 for 1:1, 2 for 4:3, 3 for 16:9
												// and 4 for 2.21:1
} ;




#endif